@@ -3,7 +3,13 @@
 from __future__ import absolute_import
 from thor.http.client import HttpClient
 from thor.http.server import HttpServer
-from thor.http.common import header_names, header_dict, get_header, \
-  safe_methods, idempotent_methods, hop_by_hop_hdrs
+from thor.http.common import (
+    header_names,
+    header_dict,
+    get_header,
+    safe_methods,
+    idempotent_methods,
+    hop_by_hop_hdrs,
+)
 
 debug = False
